import csv
import json
import sqlite3
import hashlib
import asyncio
from dataclasses import dataclass
import networkx
from structlog import get_logger

from .beakers import Beaker, SqliteBeaker, TempBeaker

log = get_logger()

"""
Implementation thoughts

Originally beaker was the central object, but pretty quickly
most functionality was moved to recipe. Now beaker is just a
thin wrapper around a sqlite table.

Recipe is the central object. It contains a list of pours,
which are the instructions for how to transform data from one
beaker to another.

Recipe also contains a list of beakers, which are just sqlite
tables. Beakers are created by the recipe, and are passed to
pours as arguments.

Beakers right now know they are stored in SQLite, but I think
that will be abstracted away. Beakers should be able to be
stored in any database, or on disk.

Core functionality of a beaker:
- initialize (create table if it doesn't exist)
- add item with optional link to another item
    - should this be links? (probably)
- iterate over items
- get item by id
- count items

Ultimately, this is a graph problem. Beakers are nodes, and
pours are edges. The recipe is the graph.
"""


def get_sha512(filename: str) -> str:
    with open(filename, "rb") as file:
        return hashlib.sha512(file.read()).hexdigest()


@dataclass(eq=True, frozen=True)
class Pour:
    """
    A pour is an edge in the graph.

    It contains a function that transforms data
    from one beaker to another.
    """

    transform_func: callable


@dataclass(eq=True, frozen=True)
class Conditional:
    """
    A conditional is a decision point in the graph.

    condition_func should return values, and the
    values will be used to determine which path
    to take.
    """

    condition_func: callable


class Recipe:
    def __init__(self, name):
        self.name = name
        self.graph = networkx.DiGraph()
        self.beakers = {}
        self.db = sqlite3.connect("beakers.db")
        self.cursor = self.db.cursor()
        self.cursor.row_factory = sqlite3.Row
        self.cursor.execute(
            "CREATE TABLE IF NOT EXISTS _metadata (table_name TEXT PRIMARY KEY, data JSON)"
        )

    def __repr__(self) -> str:
        return f"Recipe({self.name})"

    def add_beaker(self, name: str, temp: bool = False) -> Beaker:
        self.graph.add_node(name)
        if temp:
            self.beakers[name] = TempBeaker(name, self)
        else:
            self.beakers[name] = SqliteBeaker(name, self)
        return self.beakers[name]

    def add_pour(
        self, from_beaker: str, to_beaker: str, transform_func: callable
    ) -> None:
        self.graph.add_edge(from_beaker, to_beaker, transform=transform_func)

    def add_conditional(
        self,
        from_beaker: str,
        condition_func: callable,
        if_true: str,
        if_false: str,
    ) -> None:
        # first add a transform to evaluate the conditional
        cond_name = f"cond-{from_beaker}-{condition_func.__name__}"
        self.add_beaker(cond_name, temp=True)
        self.add_pour(
            from_beaker,
            cond_name,
            lambda data: (data, condition_func(data)),
        )

        # then add two filtered paths that remove the condition result
        self.graph.add_edge(
            cond_name,
            if_true,
            filter_func=lambda data, condition: data if condition else None,
        )
        self.graph.add_edge(
            cond_name,
            if_false,
            filter_func=lambda data, condition: data if not condition else None,
        )

    def get_metadata(self, table_name) -> dict:
        self.cursor.execute(
            "SELECT data FROM _metadata WHERE table_name = ?",
            (table_name,),
        )
        try:
            data = self.cursor.fetchone()["data"]
            log.debug("get_metadata", table_name=table_name, data=data)
            return json.loads(data)
        except TypeError:
            log.debug("get_metadata", table_name=table_name, data={})
            return {}

    def save_metadata(self, table_name: str, data: dict) -> None:
        data_json = json.dumps(data)
        log.info("save_metadata", table_name=table_name, data=data_json)
        # sqlite upsert
        self.cursor.execute(
            "INSERT INTO _metadata (table_name, data) VALUES (?, ?) ON CONFLICT(table_name) DO UPDATE SET data = ?",
            (table_name, data_json, data_json),
        )
        self.db.commit()

    def csv_to_beaker(self, filename: str, beaker_name: str) -> None:
        beaker = self.beakers[beaker_name]
        lg = log.bind(beaker=beaker, filename=filename)
        # three cases: empty, match, mismatch
        # case 1: empty
        if len(beaker) == 0:
            with open(filename, "r") as file:
                reader = csv.DictReader(file)
                added = 0
                for row in reader:
                    beaker.add_item(row)
                    added += 1
            lg.info("from_csv", case="empty", added=added)
            meta = self.get_metadata(beaker.name)
            meta["sha512"] = get_sha512(filename)
            self.save_metadata(beaker.name, meta)
        else:
            old_sha = self.get_metadata(beaker.name).get("sha512")
            new_sha = get_sha512(filename)
            if old_sha != new_sha:
                # case 3: mismatch
                lg.info("from_csv", case="mismatch", old_sha=old_sha, new_sha=new_sha)
                raise Exception("sha512 mismatch")
            else:
                # case 2: match
                lg.info("from_csv", case="match")
        return beaker

    def solve_dag(self):
        """
        Solve the DAG by topological sort.
        """
        for node in networkx.topological_sort(self.graph):
            print(node)

    def run_linearly(self):
        log.info("recipe", recipe=self)
        loop = asyncio.get_event_loop()

        for pour in self.pours:
            log.info(
                "pour",
                from_beaker=pour.from_beaker,
                to_beaker=pour.to_beaker,
                to_pour=len(pour.from_beaker),
            )
            for id, item in pour.from_beaker.items():
                log.info("pour_item", id=id, item=item)
                transformed = loop.run_until_complete(pour.transform(item))
                pour.to_beaker.add_item(transformed, pour.from_beaker.name, id)
